package JHeroic;

import java.net.*;
import java.io.*;
import java.util.ArrayList;

/**
 *
 * @author Szymon Bobek
 * @author Michal Gawedzki
 * @version 1.0.1
 *
 * Implementation of JHeroicInterface, allows to communicate with HeaRT using
 * HeaRT protocol in any Java application.
 *
 * 17.06.2009
 */
public class JHeroic implements JHeroicInterface
{
    /**
     * Name of host, where HeaRT server is working
     */
    public String hostName = null;

    /**
     * Debug mode on (1)/off (0)
     */
    public int debugMode = 0;

    /**
     * ipAddress of host, where HeaRT server is working
     */
    public InetAddress ipAddress = null;

    /**
     * port where HeaRT server is working
     */
    public int port = 0;
    
    /**
     * JHeroic constructor. Set debugMode to 1 if you want to save all exceptions to debug file, otherwise set 0.
     * Set hostName to name of the host and port to HeaRT port. This constructor does not establish any connection.
     *
     * @param debugMode 1 = Debug mode on, 0 = Debug mode off
     * @param hostName Name of the host, where application can connect do HeaRT server
     * @param port HeaRT server working port
     */
    public JHeroic(int debugMode, String hostName, int port)
    {
        this.debugMode = debugMode;
        this.hostName = hostName;
        this.port = port;
    }

    /**
     * JHeroic constructor. Set debugMode to 1 if you want to save all exceptions to debug file, otherwise set 0.
     * Set ipAddress to ip of the host and port to HeaRT port. This constructor does not establish any connection.
     *
     * @param debugMode 1 = Debug mode on, 0 = Debug mode off
     * @param ipAddress IP Address of HeaRT server
     * @param port HeaRT server working port
     */
    public JHeroic(int debugMode, InetAddress ipAddress, int port)
    {
        this.debugMode = debugMode;
        this.ipAddress = ipAddress;
        this.port = port;
    }

    /**
     * Returns list of all models in HeaRT as JHModel's array
     *
     * @return list of all models in HeaRT
     * @throws java.lang.Exception
     */
    public ArrayList<JHModel> getModelList() throws Exception
    {
        ArrayList<JHModel> modelList = new ArrayList<JHModel>();
        String request = "[model,getlist].";
        String replay = this.performRequest(request);

        if(replay.compareTo("[]") == 0)
            return modelList;
     
        replay.trim();
        replay = replay.substring(1, replay.length()-1);
        replay.trim();
        if(replay.compareTo("") == 0 || replay == null)
            return modelList;

        String[] models = replay.split(",");

        for(int i = 0; i < models.length;i++)
        {
            models[i] = models[i].trim();
            models[i+1] = models[i+1].trim();
            String userName = models[i];
            String modelName = models[i+1];

            char bChar = userName.charAt(0);
            char eChar = userName.charAt(userName.length()-1);

            if(bChar == '[')
                userName = userName.substring(1,userName.length());
            if(eChar == ']')
                userName = userName.substring(0,userName.length()-1);

            char bCharM = modelName.charAt(0);
            char eCharM = modelName.charAt(modelName.length()-1);

            if(bCharM == '[')
                modelName = modelName.substring(1,modelName.length());
            if(eCharM == ']')
                modelName = modelName.substring(0,modelName.length()-1);

            JHModel model = new JHModel(userName,modelName);
            modelList.add(model);
            i++;
        }

        return modelList;
    }

    /**
     * Returns user specified model as String
     *
     * @param userName Name of the user
     * @param modelName Name of the model
     * @return model as string without [ ] chars
     * @throws java.lang.Exception
     */
    public String getUserModel(String userName, String modelName) throws Exception
    {
        String request = "[model,get,'" + modelName + "','" + userName + "'].";
        String replay = this.performRequest(request);
        replay = replay.substring(1, replay.length()-1);
        return replay;
    }

    /**
     * Adds new model for specified user
     *
     * @param userName Name of the user
     * @param modelName Name of the model
     * @param model model
     * @return request status
     * @throws java.lang.Exception
     */
    public String addUserModel(String userName, String modelName, String model) throws Exception
    {
        String r = java.util.regex.Matcher.quoteReplacement("\\'");
        model = model.replaceAll("'",r);
        String request = "[model,add,'" + modelName + "','" + userName + "','" + model + "'].";
        String replay = this.performRequest(request);
        return replay;
    }

    /**
     * Removes user specified model
     *
     * @param userName Name of the user
     * @param modelName Name of the model
     * @return request status
     * @throws java.lang.Exception
     */
    public String removeUserModel(String userName, String modelName) throws Exception
    {
        String request = "[model,remove,'" + modelName + "','" + userName + "'].";
        String replay = this.performRequest(request);
        return replay;
    }

    /**
     * This method runs HeaRT with given model, inference type, tables and state
     * 
     * @param userName name of the model owner
     * @param modelName model name
     * @param infType inference type, "ddi" = Data Driven, "gdi" = Goal Driven, "tdi" = Token Driven, any other string = Data Driven (default)
     * @param tables list of string, one string = one table
     * @param state name or full definition of state
     * @return request status
     * @throws java.lang.Exception
     */
    public String runInference(String userName, String modelName, String infType, ArrayList<String> tables, String state) throws Exception
    {
        String iType = null;
        String tabStr = null;

        if(infType.compareTo("ddi") == 0)
        {
            iType = "dd";
        }
        else if(infType.compareTo("gdi") == 0)
        {
            iType = "gdi";
        }
        else if(infType.compareTo("tdi") == 0)
        {
            iType = "td";
        }
        else
        {
            iType = "dd";
        }

        for(int i = 0; i < tables.size(); i++)
        {
            if(i == 0)
            {
                tabStr = tables.get(i);
                continue;
            }
            
            tabStr = tabStr + "," + tables.get(i);
        }

        String request = "[model,run,'" + modelName + "','" + userName + "','" + iType + "','[" + tabStr + "]','" + state + "'].";
        String replay = this.performRequest(request);
        return replay;
    }

    /**
     * This method adds new state to given model
     *
     * @param userName model owner name
     * @param modelName model name
     * @param stateName new state name
     * @param stateDef new state definition
     * @return request status
     * @throws java.lang.Exception
     */
    public String addStateToModel(String userName, String modelName, String stateName, String stateDef) throws Exception
    {
        String request = "[state,add,'" + modelName + "','" + userName + "','" + stateName + "','" + stateDef + "'].";
        String replay = this.performRequest(request);
        return replay;
    }

    /**
     * This method removes given state from given model
     *
     * @param userName model owner name
     * @param modelName model name
     * @param stateName name of state do delete
     * @return request status
     * @throws java.lang.Exception
     */
    public String removeStateFromModel(String userName, String modelName, String stateName) throws Exception
    {
        String request = "[stste,remove,'" + modelName + "','" + userName + "','" + stateName + "'].";
        String replay = this.performRequest(request);
        return replay;
    }

    /**
     * Returns current protocol version
     *
     * @return protocol version
     */
    public String getProtocolVersion()
    {
        return protocolVersion;
    }

    /**
     * Allows to run model verification in HeaRT engine
     * @param userName model's owner
     * @param modelName model name
     * @param mode mode of verification, "comp" = vcomplete, "contr" = vcontradict, "subs" = vsubsume, "redu" = vreduce
     * @return verification status
     * @throws java.lang.Exception
     */
    public String verifyModel(String userName, String modelName, String mode) throws Exception
    {
        String vType = null;

        if(mode.compareTo("comp") == 0)
            vType = "vcomplete";
        else if(mode.compareTo("contr") == 0)
            vType = "vcontradict";
        else if(mode.compareTo("subs") == 0)
            vType = "vsubsume";
        else if(mode.compareTo("redu") == 0)
            vType = "vreduce";
         
        String request = "[model,verify,'" + vType + "','" + modelName + "','" + userName + "'].";
        String replay  = this.performRequest(request);
        return replay;
    }

    /**
     * Internal connection method
     *
     * @param request
     * @return
     * @throws java.lang.Exception
     */
    private String performRequest(String request) throws Exception
    {
        PrintWriter out = null;
        BufferedReader in = null;
        Socket socket = null;
        String replay = null;

        try {

            if(this.hostName == null && this.ipAddress != null)
            {
                socket = new Socket(this.ipAddress,this.port);
            }
            else if(this.ipAddress == null && this.hostName != null)
            {
                socket = new Socket(this.hostName,this.port);
            }

            out = new PrintWriter(socket.getOutputStream(), true);
            in = new BufferedReader(new InputStreamReader(socket.getInputStream()));

            out.println(request);
            out.flush();

            replay  = in.readLine();
        }
        catch(Exception e)
        {
            if(this.debugMode == 1)
            {
                FileWriter outFile = new FileWriter("debug.log");
                PrintWriter fileWriter = new PrintWriter(outFile);

                fileWriter.println(e.getMessage());
                e.printStackTrace(fileWriter);
                fileWriter.close();
                outFile.close();

                System.err.println("Error data saved to debug.log");
                throw e;
            }
            else
            {
                throw e;
            }
        }
        finally
        {
            return replay;
        }
    }
}
